/* rover.c */ 
#include "kernel.h"
#include "kernel_id.h"
#include "ecrobot_interface.h"

#include "string.h"
#include "stdio.h"
/* own definitions */
#include "config.h"


#define ROVER_SPEED -50
#define MAX_SPEED -70
#define UPDATE_INTERVAL 500
#define MAX_ADJUST_STEERING_MOTOR_COUNT 35


char PIN[] = "1234";

/* global variables  */
U8 is_running = 0;

S8 motor_update_flag = 0; 
S32 previous_left_speed = 0;
S32 previous_right_speed = 0;
S32 current_left_speed = 0;
S32 current_right_speed = 0;

U16 current_light_value = 0;
U32 light_value_sum = 0;
U32 light_count = 0;

S32 current_sonar_value = -1;
S32 previous_sonar_value = -1;

U32 adjust_steering_speed_count = 0;
U32 reverse_adjust_steering_speed_count = 0;
S32 current_steering_speed = 0;
S32 previous_steering_speed = 0;

int is_connected = 0;
U32 rev_msg_len = 0;
char rev_buf[32];
int bt_detected_pos = 0; // -1 means left, 1 means right
int bt_detected_orientation = 0; // 1 means clockwise, -1 means anti-clockwise
int bt_detected_flag = 0;

U32 front_touch_count = 0;
U32 rear_touch_count = 0;
U8 current_front_touch_value = 0;
U8 current_rear_touch_value = 0;

/* OSEK declarations */
DeclareResource(MotorValueRes);
DeclareResource(BTValueRes);
DeclareCounter(SysTimerCnt);
DeclareTask(MotorTask);
DeclareTask(LightTask);
DeclareTask(LCDTask);
DeclareTask(BTTask);

/* LEJOS OSEK hooks */
void ecrobot_device_initialize()
{
	ecrobot_init_bt_slave(PIN);
	ecrobot_init_nxtcolorsensor(LIGHT_PORT, NXT_COLORSENSOR);
	nxt_motor_set_speed(RIGHT_MOTOR, 0, 1); 
	nxt_motor_set_speed(LEFT_MOTOR, 0, 1);
	nxt_motor_set_speed(STEERING_MOTOR, 0, 1);
}	

void ecrobot_device_terminate()
{
	nxt_motor_set_speed(RIGHT_MOTOR, 0, 1); 
	nxt_motor_set_speed(LEFT_MOTOR, 0, 1); 
	nxt_motor_set_speed(STEERING_MOTOR, 0, 1);
	ecrobot_term_nxtcolorsensor(LIGHT_PORT);
	ecrobot_term_bt_connection();
}

/* LEJOS OSEK hook to be invoked from an ISR in category 2 */
void user_1ms_isr_type2(void)
{
	StatusType ercd;

	ercd = SignalCounter(SysTimerCnt); /* Increment OSEK Alarm Counter */ 
	if(ercd != E_OK)
	{
		ShutdownOS(ercd);
	}
}


U16 get_current_light_value()
{
	
	ecrobot_process_bg_nxtcolorsensor();
	ecrobot_set_nxtcolorsensor(LIGHT_PORT, NXT_LIGHTSENSOR_RED);
	return ecrobot_get_nxtcolorsensor_light(LIGHT_PORT);
}

int wait_for_connection(char* passkey, U32 duration)
{
	int is_connected = 0;
        U32 start = systick_get_ms();
        while(is_connected != 1
                  && (duration == 0 || (systick_get_ms() - start) <= duration))
        {
                ecrobot_init_bt_slave(passkey);

                if (ecrobot_get_bt_status() == BT_STREAM)
                {
                        is_connected = 1;
                        return is_connected;
                }
        }
        return is_connected;
}



/*============================================================================
 * Task: MotorTask
 */


TASK(MotorTask)
{
	if(!is_running){
		TerminateTask();
	}
	

	S32 l_speed = 0;
	S32 r_speed = 0;
	S32 s_speed = 0;
	int update_flag = 0;
	
	GetResource(MotorValueRes);
	l_speed = current_left_speed;
	r_speed = current_right_speed;
	s_speed = current_steering_speed;
	update_flag = motor_update_flag;
	ReleaseResource(MotorValueRes);

	nxt_motor_set_speed(STEERING_MOTOR, s_speed, 1);

	
	if(update_flag == 0){
		// we allow updates from the BTTask only if no updates coming from the LightTask
		int flag = 0;
		int pos = 0;
		GetResource(BTValueRes);
		flag = bt_detected_flag;
		pos = bt_detected_pos;
		ReleaseResource(BTValueRes);	
		if(flag != 0){
			if(pos == -1){
				l_speed /= 2;
			}else if(pos == 1){
				r_speed /= 2;
			}
		}
	}
	

	nxt_motor_set_speed(LEFT_MOTOR, l_speed, 1);
	nxt_motor_set_speed(RIGHT_MOTOR, r_speed, 1);

	TerminateTask();
}


/*============================================================================
 * Task: LightTask
 */
TASK(LightTask)
{
	
	if(!is_running){
		TerminateTask();
	}
	
	current_light_value = get_current_light_value();
	S32 lr_speed = -MAX_SPEED;
	S32 s_speed = STEERING_LIMIT;
	
	int flag = 0;
	int orientation = 0;
	GetResource(BTValueRes);
	flag = bt_detected_flag;
	orientation = bt_detected_orientation;
	ReleaseResource(BTValueRes);	
	if(flag != 0){
		if(orientation == 1){
			s_speed = -STEERING_LIMIT;
		}
	}

	U32 adjust_threshold = MAX_ADJUST_STEERING_MOTOR_COUNT;
	U32 reverse_adjust_threshold = MAX_ADJUST_STEERING_MOTOR_COUNT;

	GetResource(MotorValueRes);
	if(current_light_value < BLUE_LIGHT_THRESHOLD) {
		// touch the blue line
		if(motor_update_flag == 0){
			adjust_steering_speed_count = 0;
			reverse_adjust_steering_speed_count = 0;
			motor_update_flag = 1;
		}
	}

	if(motor_update_flag == 1){
		if(adjust_steering_speed_count < adjust_threshold){
			previous_steering_speed = s_speed; 
			adjust_steering_speed_count += 1;
		}else{
			motor_update_flag += 1;
		}
	}else if(motor_update_flag == 2){
		if(reverse_adjust_steering_speed_count < reverse_adjust_threshold){
			reverse_adjust_steering_speed_count += 1;
		}else{
			motor_update_flag = 0;
			lr_speed = ROVER_SPEED;
			previous_steering_speed = 0;
		}
		s_speed = -previous_steering_speed;	
	}else {
		lr_speed = ROVER_SPEED;		
		s_speed = 0;
	}	

	current_left_speed = lr_speed;
	current_right_speed = lr_speed;
	current_steering_speed = s_speed;
	ReleaseResource(MotorValueRes);	
	

	TerminateTask();
}


/* Sub functions */
S32 FrictionComp(S32 ratio, S32 offset)
{
  if (ratio > 0)
  {
    return ((100-offset)*ratio/100 + offset);
  }
  else if (ratio < 0)
  {
    return ((100-offset)*ratio/100 - offset);
  }
  else
  {
    return ratio;
  }
}

/*============================================================================
 * Task: LCDTask
 */
TASK(LCDTask)
{
	/*
	if(!is_running){
		TerminateTask();
	}
	
	
	display_clear(0);

	display_goto_xy(0, 1);
	display_string("Color: ");
	display_int(current_light_value, 0);

	display_goto_xy(0, 2);
	display_string("BT status: ");
	display_int(ecrobot_get_bt_status(), 0);

	char* p = "No rev!";
	char* q = "";
	int flag = 0;
	int pos = 0;
	GetResource(BTValueRes);
	flag = bt_detected_flag;
	pos = bt_detected_pos;
	ReleaseResource(BTValueRes);	
	if(flag != 0){
		p = "Rev: ";
		q = "M";
		if(pos == -1){
			q = "L";
		}else if(pos == 1){
			q = "R";
		}
		display_goto_xy(0, 3);
		display_string(p);
		display_string(q);

		display_goto_xy(0, 4);
		display_string("buf size: ");
		display_int(strlen(p), 0);
	}

	
	
	display_update();
	*/
	TerminateTask();
	
}


/*============================================================================
 * Task: BTTask
 */

TASK(BTTask)
{

	if(is_connected == 0){
		display_clear(0);
		display_goto_xy(0, 1);
		display_string("Waiting...");
		display_update();
		is_connected = wait_for_connection(PIN, 200);
	}
	
	if(is_connected == 0){			
		TerminateTask();
	}

	if(!is_running){
		current_left_speed = 0;
		current_right_speed = 0;
		current_steering_speed = 0;
		motor_update_flag = 0;
		previous_steering_speed = 0;
		adjust_steering_speed_count = 0;
		reverse_adjust_steering_speed_count = 0;
		is_running = 1;		
	}

	U32 start = systick_get_ms();	
	rev_msg_len = 0;
	memset(rev_buf, 0, sizeof(rev_buf));
	while(rev_msg_len == 0 && (systick_get_ms() - start) <= 50){
		rev_msg_len = ecrobot_read_bt(rev_buf, 0, 32);
	}
	
	
	if(rev_msg_len > 0){
		display_clear(0);

		display_goto_xy(0, 1);
		display_string("Color: ");
		display_int(current_light_value, 0);

		display_goto_xy(0, 2);
		display_string("BT status: ");
		display_int(ecrobot_get_bt_status(), 0);
		
		display_goto_xy(0, 3);
		display_string(rev_buf);

		display_goto_xy(0, 4);
		display_string("buf size: ");
		display_int(rev_msg_len, 0);
		
		display_update();
	}
	
	
	GetResource(BTValueRes);
	if(rev_msg_len > 0){
		bt_detected_flag = 1;
		bt_detected_pos = 0;
		bt_detected_orientation = 0;
		if(rev_buf[0] == 'L'){
			bt_detected_pos = -1;			
		}else if(rev_buf[0] == 'R'){
			bt_detected_pos = 1;			
		}
		if(rev_msg_len == 1){
			if(rev_buf[0] == 'N'){
				bt_detected_orientation = -1;			
			}else if(rev_buf[0] == 'P'){
				bt_detected_orientation = 1;			
			}			
		}else if(rev_msg_len == 2){
			if(rev_buf[1] == 'N'){
				bt_detected_orientation = -1;			
			}else if(rev_buf[1] == 'P'){
				bt_detected_orientation = 1;			
			}
		}
	}else{
		bt_detected_flag = 0;
	}
	ReleaseResource(BTValueRes);
	
	TerminateTask();

/*	static SINT bt_status = BT_NO_INIT;
  	while(1)
  	{
		ecrobot_init_bt_slave("LEJOS-OSEK");

    		if (ecrobot_get_bt_status() == BT_STREAM && bt_status != BT_STREAM)
    		{
	     	 	display_clear(0);
	      		display_goto_xy(0, 0);
	      		display_string("[BT]");
			display_int(bt_status, 0);
	      		display_update();
    		}
    		bt_status = ecrobot_get_bt_status();
  	}*/
}
/******************************** END OF FILE ********************************/

